@@ -38,7 +38,7 @@ static char copyright[] =
 #endif /* not lint */
 
 #ifndef lint
-static char sccsid[] = "@(#)rmail.c	8.1 (Berkeley) 5/31/93";
+static char sccsid[] = "@(#)rmail.c	8.3 (Berkeley) 5/15/95";
 #endif /* not lint */
 
 /*
@@ -78,6 +78,10 @@ static char sccsid[] = "@(#)rmail.c	8.1 (Berkeley) 5/31/93";
 #include <sysexits.h>
 #include <unistd.h>
 
+#ifndef MAX
+# define MAX(a, b)	((a) < (b) ? (b) : (a))
+#endif
+
 void err __P((int, const char *, ...));
 void usage __P((void));
 
@@ -193,9 +197,7 @@ main(argc, argv)
 					err(EX_TEMPFAIL, NULL);
 			}
 			if (fplen + len + 2 > fptlen) {
-				fptlen += fplen + len + 2;
-				if ( fptlen < 256 )	
-					fptlen = 256;
+				fptlen += MAX(fplen + len + 2, 256);
 				if ((from_path =
 				    realloc(from_path, fptlen)) == NULL)
 					err(EX_TEMPFAIL, NULL);
@@ -209,6 +211,8 @@ main(argc, argv)
 		/* Save off from user's address; the last one wins. */
 		for (p = addrp; *p && !isspace(*p); ++p);
 		*p = '\0';
+		if (*addrp == '\0')
+			addrp = "<>";
 		if (from_user != NULL)
 			free(from_user);
 		if ((from_user = strdup(addrp)) == NULL)
@@ -231,54 +235,42 @@ main(argc, argv)
 	args[i++] = "-odq";		/* Queue it, don't try to deliver. */
 	args[i++] = "-oi";		/* Ignore '.' on a line by itself. */
 
-	if (from_sys != NULL) {		/* Set sender's host name. */
-		if (strchr(from_sys, '.') == NULL) {
-			/*
-			(void)snprintf(buf, sizeof(buf),
-			    "-oMs%s.%s", from_sys, domain);
-			 */
-			strncpy(buf, "-oMs", sizeof(buf));
-			strncat(buf, from_sys, sizeof(buf));
-			strncat(buf, domain, sizeof(buf));
-		}
-		else {
-			/*
-			(void)snprintf(buf, sizeof(buf), "-oMs%s", from_sys);
-			 */
-			strncpy(buf, "-oMs", sizeof(buf) - 1);
-			strncat(buf, from_sys, sizeof(buf) - 1 - strlen(buf));
-		}
-		if ((args[i++] = strdup(buf)) == NULL)
-			 err(EX_TEMPFAIL, NULL);
-	}
-					/* Set protocol used. */
-	/* (void)snprintf(buf, sizeof(buf), "-oMr%s", domain); */
-	strncpy(buf, "-oMr", sizeof(buf));
-	strncat(buf, domain, sizeof(buf));
+	/* set from system and protocol used */
+	if (from_sys == NULL)
+		(void)snprintf(buf, sizeof(buf), "-p%s", domain);
+	else if (strchr(from_sys, '.') == NULL)
+		(void)snprintf(buf, sizeof(buf), "-p%s:%s.%s",
+			domain, from_sys, domain);
+	else
+		(void)snprintf(buf, sizeof(buf), "-p%s:%s", domain, from_sys);
 	if ((args[i++] = strdup(buf)) == NULL)
 		err(EX_TEMPFAIL, NULL);
 
 					/* Set name of ``from'' person. */
-	/*
 	(void)snprintf(buf, sizeof(buf), "-f%s%s",
 	    from_path ? from_path : "", from_user);
-	 */
-	strncpy(buf, "-f", sizeof(buf));
-	if ( from_path )
-		strncat(buf, from_path, sizeof(buf));
-	strncat(buf, from_user, sizeof(buf));
-
 	if ((args[i++] = strdup(buf)) == NULL)
 		err(EX_TEMPFAIL, NULL);
 
 	/*
 	 * Don't copy arguments beginning with - as they will be
 	 * passed to sendmail and could be interpreted as flags.
+	 * To prevent confusion of sendmail wrap < and > around
+	 * the address (helps to pass addrs like @gw1,@gw2:aa@bb)
 	 */
-	do {
-		if (*argv && **argv == '-')
+	while (*argv) {
+		if (**argv == '-')
 			err(EX_USAGE, "dash precedes argument: %s", *argv);
-	} while ((args[i++] = *argv++) != NULL);
+		if (strchr(*argv, ',') == NULL || strchr(*argv, '<') != NULL)
+			args[i++] = *argv;
+		else {
+			if ((args[i] = malloc(strlen(*argv) + 3)) == NULL)
+				err(EX_TEMPFAIL, "Cannot malloc");
+			sprintf (args [i++], "<%s>", *argv);
+		}
+		argv++;
+	} 
+	args[i] = 0;
 
 	if (debug) {
 		(void)fprintf(stderr, "Sendmail arguments:\n");
